@@ -3,15 +3,15 @@ After ingesting the XML documents (configuration and request) the module generat
 1. A collection of records
 2. Index
 """
-import xml, string, collections
+import xml, collections
 import xml.dom
 import xml.dom.minidom
 import re, shelve, os, sys
 try:
-  from __init__ import DOC_DIR, version, PACKAGE_DIR
+  from __init__ import DOC_DIR, version, PACKAGE_DIR, VERSION_DIR
   import utilities
 except:
-  from dreqPy.__init__ import DOC_DIR, version, PACKAGE_DIR
+  from dreqPy.__init__ import DOC_DIR, version, PACKAGE_DIR, VERSION_DIR
   from dreqPy import utilities
 
 python2 = True
@@ -33,13 +33,26 @@ charmeTempl = """<span title="Using the CHARMe annotation system">Comment on thi
 </span>
 """
 
+dreqMonitoring = '''<!-- Global site tag (gtag.js) - Google Analytics -->
+<script async src="https://www.googletagmanager.com/gtag/js?id=UA-126903002-1"></script>
+<script>
+  window.dataLayer = window.dataLayer || [];
+  function gtag(){dataLayer.push(arguments);}
+  gtag('js', new Date());
+
+  gtag('config', 'UA-126903002-1');
+</script>
+'''
+
+
 jsh='''<link type="text/css" href="/css/jquery-ui-1.8.16.custom.css" rel="Stylesheet" />
  <script src="/js/2013/jquery.min.js" type="text/javascript"></script>
  <script src="/js/2013/jquery-ui.min.js" type="text/javascript"></script>
  <script src="/js/2013/jquery.cookie.js" type="text/javascript"></script>
+%s
 
 <link type="text/css" href="/css/dreq.css" rel="Stylesheet" />
-'''
+''' % dreqMonitoring
 
 def dref(i,x):
   return i._inx.uid[i.__dict__[x]]
@@ -82,8 +95,6 @@ object._h: a python named tuple describing the section. E.g. object._h.title is
      dreqItem.ns = ns
      return dreqItem
 
-         
-
 ##
 ## from http://stackoverflow.com/questions/4474754/how-to-keep-comments-while-parsing-xml-using-python-elementtree
 ##
@@ -153,6 +164,7 @@ class dreqItemBase(object):
        _linkAttrStyle = {}
        __charmeEnable__ = {}
        _extraHtml = {}
+       _strictRequired = False
 
        def __init__(self,idict=None,xmlMiniDom=None,id='defaultId',etree=False):
          self._strictRead = True
@@ -168,7 +180,11 @@ class dreqItemBase(object):
          if dictMode:
            self.dictInit( idict )
          elif mdMode:
-           self.mdInit( xmlMiniDom, etree=etree )
+           try:
+             self.mdInit( xmlMiniDom, etree=etree )
+           except:
+             print ('Exception raise initialising data request item')
+             raise
 
        def hasattr(self,tag):
          """Return True id the argument passed is the name of an attribute of this instance."""
@@ -368,19 +384,29 @@ class dreqItemBase(object):
 
        def mdInit( self, el, etree=False ):
          __doc__ = """Initialisation from a mindom XML element. The list of attributes must be set by the class factory before the class is initialised"""
-         deferredHandling=False
-         nw1 = 0
          tvtl = []
          if etree:
            ks = set( el.keys() )
            for a in self._a.keys():
              if a in ks:
                aa = '%s%s' % (self.ns,a)
-               tvtl.append( (a,True, str( el.get( a ) ) ) )
+#
+# exception trapping to provide more diagnostics when a character code issue arises
+# added in 01.00.32beta
+#
+               this = el.get( a )
+               try:
+                 tvtl.append( (a,True, str( this ) ) )
+               except:
+                 print( 'ERROR: FAILED to str %s: %s' % (a,this) )
+                 tvtl.append( (a,True, this.encode( 'utf-8' ) ) )
+
              elif self._a[a].__dict__.get( 'required', True ) in [False,'false',u'false']:
                tvtl.append( (a,True,None) )
              else:
                tvtl.append( (a,False,None) )
+               if self._strictRequired:
+                 assert False, 'Required attribute absent: %s:: %s' % (a,str(el.__dict__))
          else:
            for a in self._a.keys():
              if el.hasAttribute( a ):
@@ -390,18 +416,82 @@ class dreqItemBase(object):
 ##
              elif self._a[a].__dict__.get( 'required', True ) not in [False,'false',u'false']:
                tvtl.append( (a,False,None) )
-       
+      
+         self._tvtlInit( tvtl )
+         self._contentInitialised = True
+
+       def _tvtlInit( self, tvtl ):
+         """Coercing string values into a python objects appropriate to declared types: looping over tuples containing tags and values.
+            the code sets the default if appropriate, otherwise call _avInit to coerce the string value to an appropriate python object."""
+
+         nw1 = 0
          for a,tv,v in tvtl:
            if a == 'uid':
              uid = v
 
          for a,tv,v in tvtl:
            if tv:
-             erase = False
              if v == None:
-               pass
-               erase = True
-             elif self._a[a].type == u'xs:float':
+               self.__dict__[a] = '__tmp__'
+               delattr( self, a )
+               ### need to overwrite attribute (which is inherited from parent class) before deleting it.
+               ### this may not be needed in python3
+             else:
+               self._avInit( a, v )
+           else:
+             if a in ['uid',]:
+               thissect = '%s' % (self._h.title)
+               print ( 'ERROR.020.0001: missing uid: %s: a,tv,v: %s, %s, %s\n %s' % (thissect,a,tv,v,str( self.__dict__ )) )
+               ##if etree:
+                 ##p = self.parent_map( el )
+                 ##print ( ks )
+               raise
+                 ##import sys
+                 ##sys.exit(0)
+             self.__dict__[a] = self._d.defaults.get( a, self._d.glob )
+
+
+       def _avInit( self, a, v ):
+             """Coercing a string value into a python object appropriate to the declared type: non-numeric data types:
+                 a: the tag identifying the data element;
+                 v: the string serialisation of the data value:
+
+                 Approach
+                 --------
+                 (1) Check whether type is numric -- if true call _avInitNumeric;
+                 (2) Otherwise, deal with boolean or string list types, or default which is to take string value as is."""
+
+             numericTypes = [u'xs:float',u'aa:st__floatList', u'aa:st__floatListMonInc',u'aa:st__integerList', u'aa:st__integerListMonInc',u'xs:integer']
+
+             if self._a[a].type in numericTypes:
+               return self._avInitNumeric( a,v )
+             elif self._a[a].type == u'xs:boolean':
+               v = v in ['true','1']
+             elif self._a[a].type == u'aa:st__stringList':
+               if v.find(' ') != -1:
+                 v = tuple( v.split() )
+               else:
+                 v = (v,)
+
+             elif self._a[a].type not in [u'xs:string', u'aa:st__uid', 'aa:st__fortranType', 'aa:st__configurationType','aa:st__sliceType']:
+               print ('ERROR: Type %s not recognised [%s:%s]' % (self._a[a].type,self._h.label,a) )
+             
+             self.__dict__[a] = v
+
+       def _avInitNumeric( self, a, v ):
+             """Coercing a value into a python object appropriate to the declared type: numeric data types.
+                 a: the tag identifying the data element;
+                 v: the string serialisation of the data value:
+                 
+                 list types are split into elements;
+                 integers coerced with the python int function;
+                 floats coerced with the python float function;
+                 -- if integer coercion fails, indirect coercion is attempted via int (float (v) )
+                 -- if specified in the type, monotonicity is checked.
+                 -- if type is integer, a blank is taken as zero.
+             """
+             msg = ''
+             if self._a[a].type == u'xs:float':
                if v == '':
                  v = None
                else:
@@ -439,46 +529,15 @@ class dreqItemBase(object):
                      print ( 'WARN: input integer non-compliant: %s: %s: %s [%s] %s' % (thissect,a,v0,v,uid) )
                    except:
                      msg = 'ERROR: failed to convert integer: %s: %s: %s, %s' % (thissect,a,v,type(v))
-                     deferredHandling=True
-             elif self._a[a].type == u'xs:boolean':
-               v = v in ['true','1']
-             elif self._a[a].type == u'aa:st__stringList':
-               if v.find(' ') != -1:
-                 v = tuple( v.split() )
-               else:
-                 v = (v,)
-             elif self._a[a].type not in [u'xs:string', u'aa:st__uid', 'aa:st__fortranType', 'aa:st__configurationType','aa:st__sliceType']:
-               print ('ERROR: Type %s not recognised [%s:%s]' % (self._a[a].type,self._h.label,a) )
-
-             if erase:
-               ### need to overwrite attribute (which is inherited from parent class) before deleting it.
-               ### this may not be needed in python3
-               self.__dict__[a] = '__tmp__'
-               delattr( self, a )
+                     print ( msg )
              else:
-               self.__dict__[a] = v
-           else:
-             if a in ['uid',]:
-               thissect = '%s' % (self._h.title)
-               print ( 'ERROR.020.0001: missing uid: %s: a,tv,v: %s, %s, %s\n %s' % (thissect,a,tv,v,str( self.__dict__ )) )
-               if etree:
-                 ##p = self.parent_map( el )
-                 print ( ks )
-                 raise
-                 import sys
-                 sys.exit(0)
-             self.__dict__[a] = self._d.defaults.get( a, self._d.glob )
-
-           if deferredHandling:
-             print ( msg )
-
-         self._contentInitialised = True
-
+               raise
+             self.__dict__[a] = v
     
 class config(object):
   """Read in a vocabulary collection configuration document and a vocabulary document"""
 
-  def __init__(self, configdoc='out/dreqDefn.xml', thisdoc='../workbook/trial_20150724.xml', manifest=None, useShelve=False, strings=False,configOnly=False):
+  def __init__(self, configdoc='out/dreqDefn.xml', thisdoc='../workbook/trial_20150724.xml', manifest=None, useShelve=False, strings=False,configOnly=False,lxml=False):
     self.rc = rechecks()
     self.lu = lutilsC()
     self.silent = True
@@ -495,11 +554,14 @@ class config(object):
     self.strings = strings
     self.docl = []
 
+    self._desc = {}
+    self.__desc__ = self._desc
     self.tt0 = {}
     self.tt1 = {}
     self.ttl2 = []
     self.docs = {}
     self.version = None
+    self.useLxml = lxml
 
     if manifest != None:
       assert os.path.isfile( manifest ), 'Manifest file not found: %s' % manifest
@@ -544,20 +606,33 @@ class config(object):
 ##
     self.etree = False
     self.etree = True
+    root=None
     self.ns = None
     if not self.configOnly:
       if self.etree:
-        import xml.etree.cElementTree as cel
+        if self.useLxml:
+          import lxml
+          import lxml.etree as et
+        else:
+          import xml.etree.cElementTree as cel
       
         if not pythonPre27:
           ## this for of namespace registration not available in 2.6
           ## absence of registration means module cannot write data exactly as read.
           ##
-          cel.register_namespace('', "urn:w3id.org:cmip6.dreq.dreq:a")
-          cel.register_namespace('pav', "http://purl.org/pav/2.3")
+          if self.useLxml:
+            ##et.register_namespace('', "urn:w3id.org:cmip6.dreq.dreq:a")
+            et.register_namespace('pav', "http://purl.org/pav/2.3")
+          else:
+            cel.register_namespace('', "urn:w3id.org:cmip6.dreq.dreq:a")
+            cel.register_namespace('pav', "http://purl.org/pav/2.3")
 
         if not self.strings:
-          if python2:
+          if self.useLxml:
+            ##parser = etree.XMLParser( remove_blank_text=True )
+            ##self.contentDoc = et.parse( self.vsamp, parser  )
+            self.contentDoc = et.parse( self.vsamp  )
+          elif python2:
             parser = getParser()()
             self.contentDoc = cel.parse( self.vsamp, parser=parser )
           else:
@@ -627,6 +702,15 @@ class config(object):
 ##
     self.coll['__core__'] = self.ntf( self._t0.header, self._t0.attributes, [self.tt0[k] for k in self.tt0] )
 
+    self._read_vl(vl,tables)
+
+    for k in tables:
+      self.recordAttributeDefn[k] = tables[k]
+
+    if not self.configOnly:
+      self.__read_tables__(root,tables)
+
+  def _read_vl(self,vl,tables):
     ec = {}
     for i in self.coll['__core__'].items:
       ec[i.label] = i
@@ -642,6 +726,7 @@ class config(object):
       self.tt1[t[0].label].level = t.header.level
       self.tt1[t[0].label].uid = t.header.uid
       self.tt1[t[0].label].itemLabelMode = t.header.itemLabelMode
+      self._desc[t[0].label] = t.header.description
       self.ttl2 += [thisc.__dict__[a] for a in t.attributes]
     mil = [t[1] for t in self._t2.attributes.items()]
     self.coll['__main__'] = self.ntf( self._t2.header, self._t2.attributes, self.ttl2 )
@@ -653,11 +738,7 @@ class config(object):
         assert k in ec, 'Key %s [%s] not found' % (k,sct)
         self.coll[sct].attDefn[k] = ec[k]
 
-    for k in tables:
-      self.recordAttributeDefn[k] = tables[k]
-
-    if self.configOnly:
-      return
+  def __read_tables__(self,root,tables):
     for k in tables.keys():
       if self.etree:
         vl = root.findall( './/%s%s' % (self.ns,k) )
@@ -680,7 +761,12 @@ class config(object):
           self.tables[k] = (i,t,len(il))
         
           for i in il:
-            ii = self.tableClasses[k](xmlMiniDom=i, etree=True)
+            try:
+              ii = self.tableClasses[k](xmlMiniDom=i, etree=True)
+            except:
+              print ('Exception raises instantiating item in section %s' % k)
+              print ('At item uid=%s' % str(i) )
+              raise
             self.tableItems[k].append( ii )
         elif len(vl) > 1:
           assert False, 'not able to handle repeat sections with etree yet'
@@ -914,15 +1000,24 @@ class loadDreq(object):
   configdoc: full path to associated configuration document
   useShelve: flag to specify whether to retrieve data from cache (not implemented)
   htmlStyles: dictionary of styling directives which influence structure of html page generates by the "makeHtml" method
+  lxml [False]: if true, use python lxml package for elementree module instead of the default xml package.
 """
 
-  def __init__(self,dreqXML=None, configdoc=None, useShelve=False, htmlStyles=None, strings=False, manifest=defaultManifestPath , configOnly=False):
-    if manifest == None:
+  def __init__(self, xmlVersion=None, dreqXML=None, configdoc=None, useShelve=False, htmlStyles=None, strings=False, manifest=defaultManifestPath , configOnly=False,lxml=False):
+    self._extensions_ = {}
+    if xmlVersion != None:
+      assert os.path.isdir( VERSION_DIR ),'Version diretory %s not found;\nCreate or change environment variable DRQ_VERSION_DIR' % VERSION_DIR 
+      assert os.path.isdir( '%s/%s' % (VERSION_DIR,xmlVersion) ), 'Version %s not in %s .. download from ..' % (xmlVersion,VERSION_DIR)
+      dreqXML='%s/%s/dreq.xml' % (VERSION_DIR,xmlVersion)
+      configdoc='%s/%s/dreq2Defn.xml' % (VERSION_DIR,xmlVersion)
+      manifest = None
+    elif manifest == None:
       if dreqXML == None:
        dreqXML=defaultDreqPath
       if configdoc==None:
        configdoc=defaultConfigPath
-    self.c = config( thisdoc=dreqXML, configdoc=configdoc, useShelve=useShelve,strings=strings,manifest=manifest,configOnly=configOnly)
+    self._VERSION_DIR_ = VERSION_DIR
+    self.c = config( thisdoc=dreqXML, configdoc=configdoc, useShelve=useShelve,strings=strings,manifest=manifest,configOnly=configOnly,lxml=lxml)
     self.coll = self.c.coll
     self.version = self.c.version
     self.softwareVersion = version
@@ -945,21 +1040,22 @@ class loadDreq(object):
 
 ##    dreqItemBase._htmlStyle['__general__'] = {'addRemarks':True}
 
-      self.pageTmpl = """<html><head><title>%s</title>
+      self.pageTmpl = """<html><head><title>%%s</title>
+%%s
+<link rel="stylesheet" type="text/css" href="%%scss/dreq.css">
 %s
-<link rel="stylesheet" type="text/css" href="%scss/dreq.css">
 </head><body>
 
 <div id="top">
    <div id="corner"></div>
    CMIP6 Data Request
 </div>
-<div id="nav"><div><a href="%s" title="Home">Home</a></div></div>
+<div id="nav"><div><a href="%%s" title="Home">Home</a></div></div>
 
 <div id="section">
-%s
+%%s
 </div>
-</body></html>"""
+</body></html>"""  % dreqMonitoring
 
   def getHtmlItemStyle(self, sect):
     """Get the styling method associated with a given section."""
